// Copyright (c) 2025 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build !purego

#include "textflag.h"

// func feMul(out *Element, a *Element, b *Element)
TEXT ·feMul(SB), NOSPLIT, $0-24
	MOVV	a+8(FP), R4
	MOVV	b+16(FP), R5

	MOVV	$19, R6

	MOVV	(R4), R7	// a0
	MOVV	(R5), R8	// b0
	MOVV	8(R5), R24	// b1
	MOVV	16(R5), R25	// b2
	MOVV	24(R5), R26	// b3
	MOVV	32(R5), R27	// b4

	// r0 = a0×b0
	MULV	R7, R8, R12
	MULHVU	R7, R8, R13

	// r1 = a0×b1
	MULV	R7, R24, R14
	MULHVU	R7, R24, R15

	// r2 = a0×b2
	MULV	R7, R25, R16
	MULHVU	R7, R25, R17

	// r3 = a0×b3
	MULV	R7, R26, R18
	MULHVU	R7, R26, R19

	// r4 = a0×b4
	MULV	R7, R27, R30
	MULHVU	R7, R27, R20

	MOVV	8(R4), R7	// a1

	// r1 += a1×b0
	MULV	R7, R8, R10
	MULHVU	R7, R8, R11
	ADDV	R10, R14, R14
	SGTU	R10, R14, R9
	ADDV	R11, R15, R15
	ADDV	R9, R15, R15

	// r2 += a1×b1
	MULV	R7, R24, R10
	MULHVU	R7, R24, R11
	ADDV	R10, R16, R16
	SGTU	R10, R16, R9
	ADDV	R11, R17, R17
	ADDV	R9, R17, R17

	// r3 += a1×b2
	MULV	R7, R25, R10
	MULHVU	R7, R25, R11
	ADDV	R10, R18, R18
	SGTU	R10, R18, R9
	ADDV	R11, R19, R19
	ADDV	R9, R19, R19

	// r4 += a1×b3
	MULV	R7, R26, R10
	MULHVU	R7, R26, R11
	ADDV	R10, R30, R30
	SGTU	R10, R30, R9
	ADDV	R11, R20, R20
	ADDV	R9, R20, R20

	// r0 += 19×a1×b4
	MULV	R6, R7, R7
	MULV	R7, R27, R10
	MULHVU	R7, R27, R11
	ADDV	R10, R12, R12
	SGTU	R10, R12, R9
	ADDV	R11, R13, R13
	ADDV	R9, R13, R13

	MOVV	16(R4), R7	// a2

	// r2 += a2×b0
	MULV	R7, R8, R10
	MULHVU	R7, R8, R11
	ADDV	R10, R16, R16
	SGTU	R10, R16, R9
	ADDV	R11, R17, R17
	ADDV	R9, R17, R17

	// r3 += a2×b1
	MULV	R7, R24, R10
	MULHVU	R7, R24, R11
	ADDV	R10, R18, R18
	SGTU	R10, R18, R9
	ADDV	R11, R19, R19
	ADDV	R9, R19, R19

	// r4 += a2×b2
	MULV	R7, R25, R10
	MULHVU	R7, R25, R11
	ADDV	R10, R30, R30
	SGTU	R10, R30, R9
	ADDV	R11, R20, R20
	ADDV	R9, R20, R20

	// r0 += 19×a2×b3
	MULV	R6, R7, R7
	MULV	R7, R26, R10
	MULHVU	R7, R26, R11
	ADDV	R10, R12, R12
	SGTU	R10, R12, R9
	ADDV	R11, R13, R13
	ADDV	R9, R13, R13

	// r1 += 19×a2×b4
	MULV	R7, R27, R10
	MULHVU	R7, R27, R11
	ADDV	R10, R14, R14
	SGTU	R10, R14, R9
	ADDV	R11, R15, R15
	ADDV	R9, R15, R15

	MOVV	24(R4), R7	// a3

	// r3 += a3×b0
	MULV	R7, R8, R10
	MULHVU	R7, R8, R11
	ADDV	R10, R18, R18
	SGTU	R10, R18, R9
	ADDV	R11, R19, R19
	ADDV	R9, R19, R19

	// r4 += a3×b1
	MULV	R7, R24, R10
	MULHVU	R7, R24, R11
	ADDV	R10, R30, R30
	SGTU	R10, R30, R9
	ADDV	R11, R20, R20
	ADDV	R9, R20, R20

	// r0 += 19×a3×b2
	MULV	R6, R7, R7
	MULV	R7, R25, R10
	MULHVU	R7, R25, R11
	ADDV	R10, R12, R12
	SGTU	R10, R12, R9
	ADDV	R11, R13, R13
	ADDV	R9, R13, R13

	// r1 += 19×a3×b3
	MULV	R7, R26, R10
	MULHVU	R7, R26, R11
	ADDV	R10, R14, R14
	SGTU	R10, R14, R9
	ADDV	R11, R15, R15
	ADDV	R9, R15, R15

	// r2 += 19×a3×b4
	MULV	R7, R27, R10
	MULHVU	R7, R27, R11
	ADDV	R10, R16, R16
	SGTU	R10, R16, R9
	ADDV	R11, R17, R17
	ADDV	R9, R17, R17

	MOVV	32(R4), R7	// a4

	// r4 += a4×b0
	MULV	R7, R8, R10
	MULHVU	R7, R8, R11
	ADDV	R10, R30, R30
	SGTU	R10, R30, R9
	ADDV	R11, R20, R20
	ADDV	R9, R20, R20

	// r0 += 19×a4×b1
	MULV	R6, R7, R7
	MULV	R7, R24, R10
	MULHVU	R7, R24, R11
	ADDV	R10, R12, R12
	SGTU	R10, R12, R9
	ADDV	R11, R13, R13
	ADDV	R9, R13, R13

	// r1 += 19×a4×b2
	MULV	R7, R25, R10
	MULHVU	R7, R25, R11
	ADDV	R10, R14, R14
	SGTU	R10, R14, R9
	ADDV	R11, R15, R15
	ADDV	R9, R15, R15

	// r2 += 19×a4×b3
	MULV	R7, R26, R10
	MULHVU	R7, R26, R11
	ADDV	R10, R16, R16
	SGTU	R10, R16, R9
	ADDV	R11, R17, R17
	ADDV	R9, R17, R17

	// r3 += 19×a4×b4
	MULV	R7, R27, R10
	MULHVU	R7, R27, R11
	ADDV	R10, R18, R18
	SGTU	R10, R18, R9
	ADDV	R11, R19, R19
	ADDV	R9, R19, R19

	// First reduction chain
	MOVV	$0x0007ffffffffffff, R10
	SRLV	$51, R12, R11
	SLLV	$13, R13, R13
	OR	R11, R13, R13
	SRLV	$51, R14, R11
	SLLV	$13, R15, R15
	OR	R11, R15, R15
	SRLV	$51, R16, R11
	SLLV	$13, R17, R17
	OR	R11, R17, R17
	SRLV	$51, R18, R11
	SLLV	$13, R19, R19
	OR	R11, R19, R19
	SRLV	$51, R30, R11
	SLLV	$13, R20, R20
	OR	R11, R20, R20
	AND	R10, R12, R12
	MULV	R6, R20, R20
	AND	R10, R14, R14
	AND	R10, R16, R16
	AND	R10, R18, R18
	AND	R10, R30, R30
	ADDV	R20, R12, R12
	ADDV	R13, R14, R14
	ADDV	R15, R16, R16
	ADDV	R17, R18, R18
	ADDV	R19, R30, R30

	// Second reduction chain (carryPropagate)
	SRLV	$51, R12, R13
	SRLV	$51, R14, R15
	SRLV	$51, R16, R17
	SRLV	$51, R18, R19
	SRLV	$51, R30, R20
	AND	R10, R12, R12
	MULV	R6, R20, R20
	AND	R10, R14, R14
	AND	R10, R16, R16
	AND	R10, R18, R18
	AND	R10, R30, R30
	ADDV	R20, R12, R12
	ADDV	R13, R14, R14
	ADDV	R15, R16, R16
	ADDV	R17, R18, R18
	ADDV	R19, R30, R30

	// Store output
	MOVV	out+0(FP), R4
	MOVV	R12, (R4)
	MOVV	R14, 8(R4)
	MOVV	R16, 16(R4)
	MOVV	R18, 24(R4)
	MOVV	R30, 32(R4)
	RET

// func feSquare(out *Element, a *Element)
TEXT ·feSquare(SB), NOSPLIT, $0-16
	MOVV	a+8(FP), R4

	MOVV	$19, R5
	MOVV	$38, R6

	MOVV	(R4), R7	// l0
	MOVV	8(R4), R8	// l1
	MOVV	16(R4), R24	// l2
	MOVV	24(R4), R25	// l3
	MOVV	32(R4), R26	// l4

	// r0 = l0×l0
	MULV	R7, R7, R12
	MULHVU	R7, R7, R13

	// r1 = 2×l0×l1
	SLLV	$1, R7, R27
	MULV	R27, R8, R14
	MULHVU	R27, R8, R15

	// r2 = 2×l0×l2
	MULV	R27, R24, R16
	MULHVU	R27, R24, R17

	// r3 = 2×l0×l3
	MULV	R27, R25, R18
	MULHVU	R27, R25, R19

	// r4 = 2×l0×l4
	MULV	R27, R26, R30
	MULHVU	R27, R26, R20

	// r2 += l1×l1
	MULV	R8, R8, R10
	MULHVU	R8, R8, R11
	ADDV	R10, R16, R16
	SGTU	R10, R16, R9
	ADDV	R11, R17, R17
	ADDV	R9, R17, R17

	// r3 += 2×l1×l2
	SLLV	$1, R8, R27
	MULV	R27, R24, R10
	MULHVU	R27, R24, R11
	ADDV	R10, R18, R18
	SGTU	R10, R18, R9
	ADDV	R11, R19, R19
	ADDV	R9, R19, R19

	// r4 += 2×l1×l3
	MULV	R27, R25, R10
	MULHVU	R27, R25, R11
	ADDV	R10, R30, R30
	SGTU	R10, R30, R9
	ADDV	R11, R20, R20
	ADDV	R9, R20, R20

	// r0 += 38×l1×l4
	MULV	R5, R27, R27
	MULV	R27, R26, R10
	MULHVU	R27, R26, R11
	ADDV	R10, R12, R12
	SGTU	R10, R12, R9
	ADDV	R11, R13, R13
	ADDV	R9, R13, R13

	// r4 += l2×l2
	MULV	R24, R24, R10
	MULHVU	R24, R24, R11
	ADDV	R10, R30, R30
	SGTU	R10, R30, R9
	ADDV	R11, R20, R20
	ADDV	R9, R20, R20

	// r0 += 38×l2×l3
	MULV	R6, R24, R27
	MULV	R27, R25, R10
	MULHVU	R27, R25, R11
	ADDV	R10, R12, R12
	SGTU	R10, R12, R9
	ADDV	R11, R13, R13
	ADDV	R9, R13, R13

	// r1 += 38×l2×l4
	MULV	R27, R26, R10
	MULHVU	R27, R26, R11
	ADDV	R10, R14, R14
	SGTU	R10, R14, R9
	ADDV	R11, R15, R15
	ADDV	R9, R15, R15

	// r1 += 19×l3×l3
	MULV	R5, R25, R27
	MULV	R27, R25, R10
	MULHVU	R27, R25, R11
	ADDV	R10, R14, R14
	SGTU	R10, R14, R9
	ADDV	R11, R15, R15
	ADDV	R9, R15, R15

	// r2 += 38×l3×l4
	MULV	R6, R25, R27
	MULV	R27, R26, R10
	MULHVU	R27, R26, R11
	ADDV	R10, R16, R16
	SGTU	R10, R16, R9
	ADDV	R11, R17, R17
	ADDV	R9, R17, R17

	// r3 += 19×l4×l4
	MULV	R5, R26, R27
	MULV	R27, R26, R10
	MULHVU	R27, R26, R11
	ADDV	R10, R18, R18
	SGTU	R10, R18, R9
	ADDV	R11, R19, R19
	ADDV	R9, R19, R19

	// First reduction chain
	MOVV	$0x0007ffffffffffff, R10
	SRLV	$51, R12, R11
	SLLV	$13, R13, R13
	OR	R11, R13, R13
	SRLV	$51, R14, R11
	SLLV	$13, R15, R15
	OR	R11, R15, R15
	SRLV	$51, R16, R11
	SLLV	$13, R17, R17
	OR	R11, R17, R17
	SRLV	$51, R18, R11
	SLLV	$13, R19, R19
	OR	R11, R19, R19
	SRLV	$51, R30, R11
	SLLV	$13, R20, R20
	OR	R11, R20, R20
	AND	R10, R12, R12
	MULV	R5, R20, R20
	AND	R10, R14, R14
	AND	R10, R16, R16
	AND	R10, R18, R18
	AND	R10, R30, R30
	ADDV	R20, R12, R12
	ADDV	R13, R14, R14
	ADDV	R15, R16, R16
	ADDV	R17, R18, R18
	ADDV	R19, R30, R30

	// Second reduction chain (carryPropagate)
	SRLV	$51, R12, R13
	SRLV	$51, R14, R15
	SRLV	$51, R16, R17
	SRLV	$51, R18, R19
	SRLV	$51, R30, R20
	AND	R10, R12, R12
	MULV	R5, R20, R20
	AND	R10, R14, R14
	AND	R10, R16, R16
	AND	R10, R18, R18
	AND	R10, R30, R30
	ADDV	R20, R12, R12
	ADDV	R13, R14, R14
	ADDV	R15, R16, R16
	ADDV	R17, R18, R18
	ADDV	R19, R30, R30

	// Store output
	MOVV	out+0(FP), R4
	MOVV	R12, (R4)
	MOVV	R14, 8(R4)
	MOVV	R16, 16(R4)
	MOVV	R18, 24(R4)
	MOVV	R30, 32(R4)
	RET
